import os
from Bio.SeqIO import parse
from subprocess import call
from pandas import read_csv, DataFrame


def create_fasta(output, mearged_pike_out, dbpath):
    # Create fasta file 
    consensus = {}
    cons_conter = 0
    
    with open(f'{output}/all_consensus.fasta', 'w') as opn_fasta:
        for cons in mearged_pike_out.index:
    
            opn_fasta.write(f'>{cons_conter}\n{cons}\n')
            consensus[cons_conter] = cons
            cons_conter += 1
    
    return consensus
    
def run_blast(base, path, threads):
    
    call(f'makeblastdb -in {base} -dbtype nucl', shell=True)
    call(f'blastn -num_threads {threads}  -outfmt "7 qseqid sseqid pident evalue qcovs bitscore" -query {path}/all_consensus.fasta  -db {base} -out {path}/blast_results.txt', shell=True)

def decode_tax(base) -> dict:
    
    # DB decoder 
    # Use db header format: Kingdom    Phylum    Class    Order    Family    Genus    Species
    
    base = parse(base, 'fasta')
    taxonomy_linage = ['Kingdom', 'Phylum', 'Class', 'Order', 'Family', 'Genus', 'Species']
    tax_decoder = {}
    
    for line in base:
        
        tax_decoder[line.id] = {}
        linage = line.description.split(';')
        linage[0] = linage[0].split()[1]
    
        for i in range(len(taxonomy_linage)):
            try:
                
                tax_decoder[line.id][taxonomy_linage[i]] = linage[i]
            
            except:
                
                tax_decoder[line.id][taxonomy_linage[i]] = 'NA'
    
    return tax_decoder

def parse_blast(path, 
                base, 
                data_tax, 
                consensus, 
                identity_filter, 
                cov_lim, 
                evalue_filter):
    
    # parser of blast table
    
    blast_header = ['qseqid',
                    'sseqid', 
                    'pident',
                    'evalue',
                    'qcovs', 
                    'bitscore']
    
    blasting_results = {}
    opn_blast = read_csv(f'{path}/blast_results.txt', sep='\t', comment='#', header=None, names=blast_header)
    
    for i in opn_blast['qseqid'].unique():
        
        blast_subset = opn_blast[opn_blast["qseqid"] == i]
        blast_subset = blast_subset[blast_subset['pident'] >= identity_filter]
        blast_subset = blast_subset[blast_subset['evalue'] <= evalue_filter]
        blast_subset = blast_subset[blast_subset['qcovs'] >= cov_lim]

        blast_subset = blast_subset.sort_values(by='evalue')
        blast_subset = blast_subset.sort_values(by='pident')[::-1]

        if len(blast_subset['sseqid'].values) == 0:
            continue
            
        subject = blast_subset['sseqid'].values[0]
        blasting_results[consensus[i]] = data_tax[subject]
        
    blasting_results_df = DataFrame(blasting_results).T
    
    return blasting_results_df


def call_taxonomy(output, 
                dbpath,
                mearged_pike_out, 
                threads,
                identity_filter=95, 
                cov_lim=60, 
                evalue_filter=1e-05):

    # Creating output directory
    try:        
        os.mkdir(output)
    except FileExistsError:
        print('The output directory already exists!')
        
    consensus = create_fasta(output, mearged_pike_out, dbpath)
    run_blast(dbpath, output, threads)
    data_tax = decode_tax(dbpath)
    blasting_results_df = parse_blast(output, 
                                    dbpath, 
                                    data_tax, 
                                    consensus, 
                                    identity_filter, 
                                    cov_lim, 
                                    evalue_filter)

    return blasting_results_df




import numpy as np

def filter_cluster(clt_dat):
    
    #_____________FILTERING_BY_LENGTH_____________________________________
    median = np.median(clt_dat['Length'].values)
    st_div = np.std(clt_dat['Length'].values)
    left = median - 10
    right = median + 10
    filtered_data = clt_dat[clt_dat['Length'] > left]
    filtered_data = filtered_data[filtered_data['Length'] < right]
    #_____________________________________________________________________
    #_______________FILTERING_BY_GC_______________________________________
    median = np.median(clt_dat['GC content'].values)
    st_div = np.std(clt_dat['GC content'].values)
    left = median - 2*st_div
    right = median + 2*st_div
    filtered_data = filtered_data[filtered_data['GC content'] > left]
    filtered_data = filtered_data[filtered_data['GC content'] < right]
    #_____________________________________________________________________
    
    return  filtered_data


import os
from subprocess import call

def run_trimming(path_to_fastq, 
                 sample, 
                 output, 
                 primerF, 
                 primerR):
    
    #Trimming round 1
    command_input1 = f'{path_to_fastq}/{sample}.fastq'
    command_output1 = f'{output}/read_preprocessing/cutadapt_round1/{sample}.fastq'    
#    command1 = f'cutadapt -g ^{primerF} --minimum-length 20 --rc --discard-untrimmed -e 0.3 --quiet  -o {command_output1} {command_input1}'#add paramerts 
    command1 = f'cutadapt -g ^{primerF} --minimum-length 20 --rc -e 0.3 --discard-untrimmed --quiet  -o {command_output1} {command_input1}'#add paramerts 
    call(command1, shell=True)
    #Trimming round 2
    command_input2 = f'{output}/read_preprocessing/cutadapt_round1/{sample}.fastq'
    command_output2 = f'{output}/read_preprocessing/cutadapt_round2/{sample}.fastq'#--quality-base 20
    command2 = f'cutadapt -g ^{primerR}  --minimum-length 20 --rc -e 0.3 --discard-untrimmed --quiet -o {command_output2} {command_input2}'
    call(command2, shell=True)
    #new path to preprocessed fastq
    goodreads = f'{output}/read_preprocessing/cutadapt_round2/'

    return goodreads

def run_filtering(path_to_fastq, 
                  sample, 
                  output, 
                  minlen, 
                  maxlen, 
                  read_q_score):
    
    #Filtering
    command_input = f'{path_to_fastq}/{sample}.fastq'
    command_output = f'{output}/read_preprocessing/filtered_reads/{sample}.fastq'
    command = f'filtlong --min_length {minlen} --max_length {maxlen} --mean_q_weight {read_q_score} {command_input} > {command_output}'
    call(command, shell=True)
    #new path to preprocessed fastq
    goodreads = f'{output}/read_preprocessing/filtered_reads/'

    return goodreads



import os
import sys
import argparse 
import numpy as np
from os import listdir, mkdir
from subprocess import call
from Pike.src.get_cluster import miltiprocess_analyze
from Pike.src.get_merge import merge_output
import os
#import tensorflow as tf

def main():

    parser = argparse.ArgumentParser(description=' P I K E : metagenomic tool for noisy reads analyzing')
    group1 = parser.add_argument_group("main arguments")
    group2 = parser.add_argument_group("advance arguments")
    group3 = parser.add_argument_group("read quality arguments")
    group4 = parser.add_argument_group("clustering arguments")
    group5 = parser.add_argument_group("consensus building arguments")

    #______________________Main options_______________________________
    group1.add_argument('-fastq',
                        type=str,
                        help='Directory with merged read files',
                        required=True,
                        default=None)      
    group1.add_argument('-output',
                        type=str,
                        help='Directory with output files',
                        default='./Pike_results')
    #__________________________________________________________________
    #______________________Advance options_____________________________     
    group2.add_argument('-mode',
                        type=str,
                        help='Choice of analysis mode. single - each sample is analyzed separately, pool - joint analysis of all samples (single by default)',
                        required=True,
                        default='single')    
    group2.add_argument('-usereads', 
                        type=int,
                        help='Maximum number of reads per sample for analysis (Everything is used by default)',
                        default=np.inf)
    group2.add_argument('-threads', 
                        type=int,
                        help='Number of threads (1 by default)',
                        default=1)
    #__________________________________________________________________
    #______________________Read quality options________________________
    group3.add_argument('--trim_primer',
                        action='store_true',
                        help='Trim primers (If this mod is selected, then it is necessary to apply primer sequences)',
                        default=False)
    group3.add_argument('-primerF',
                        type=str,
                        help='Forward primer sequence',
                        default=None)   
    group3.add_argument('-primerR',
                        type=str,
                        help='Reverse primer sequence',
                        default=None)   
    group3.add_argument('-read_q_score',
                        type=int,
                        help='Median value of read quality',
                        default=20)
    group3.add_argument('-minlen',
                        type=str,
                        help='Minimum expected amplicon length (350 by default)',
                        default=350)
    group3.add_argument('-maxlen',
                        type=int,
                        help='Maximum expected amplicon length (600 by default)',
                        default=600)
    #_________________________________________________________________
    #______________________Clustering parametrs_______________________
    group4.add_argument('-umap_neighbours',
                        type=int,
                        help='Number of nearest neighbors for UMAP (30 by default)',
                        default=10)
    group4.add_argument('-cluster_size',
                        type=int,
                        help='Number of dots for HDBSCAN (30 by default)',
                        default=10)
    group4.add_argument('-k',
                        type=int,
                        help='k-mer size number (6 by default)',
                        default=6)
    #_________________________________________________________________
    #______________________Consensus parametrs________________________
    group5.add_argument('-consensus_seq_lim',
                        type=int,
                        help='Minimum number of reads supporting consensus sequence',
                        default=20)
    group5.add_argument('-letter_Q_lim',
                        type=int,
                        help='The minimum quality of the consensus letter. If the quality is less, then N is set (15 by default)',
                        default=15)
    #_________________________________________________________________
    
    if len(sys.argv)==1:
        parser.print_help(sys.stderr)
        sys.exit(1)

    args = parser.parse_args()
    #Innitializing arguments 
    #____Main options__________________________
    path_to_fastq = args.fastq
    output = args.output
    #__________________________________________
    #____Advance options_______________________
    mode = args.mode
    usereads = args.usereads #Not for pool
    threads = args.threads
    #__________________________________________
    #____Read quality options__________________
    trim_primer = args.trim_primer
    primerF = args.primerF
    primerR = args.primerR
    minlen = args.minlen
    maxlen = args.maxlen
    read_q_score = args.read_q_score
    #__________________________________________
    #____Clustering parametrs__________________
    umap_neighbours = args.umap_neighbours
    hdbscan_neighbours = args.cluster_size
    k = args.k
    #____Consensus parametrs__________________
    #__________________________________________
    consensus_seq_lim = args.consensus_seq_lim
    letter_Q_lim = args.letter_Q_lim
    #__________________________________________    


    try:        
        os.mkdir(output)
    except FileExistsError:

        print('The output directory already exists!')

    #output dir bases
    os.mkdir(f'{output}/read_preprocessing/')
    os.mkdir(f'{output}/work_dir/')
    os.mkdir(f'{output}/results/')
    #read processing dir dir bases
    if trim_primer == True:
        if primerF is None or primerR is None:
            print('For trimming, it is necessary to give primer sequences (primerF and primerR)!')
            sys.exit()

        os.mkdir(f'{output}/read_preprocessing/cutadapt_round1')
        os.mkdir(f'{output}/read_preprocessing/cutadapt_round2')

    os.mkdir(f'{output}/read_preprocessing/filtered_reads')    

    import tensorflow as tf
    tf.config.threading.set_intra_op_parallelism_threads(32) #Add in git 
   
    print('                      R U N   A N A L Y S I S                      ')
    print('===================================================================')

    miltiprocess_analyze(path_to_fastq, 
                         output,
                         mode,
                         threads,
                         usereads,
                         trim_primer,
                         primerF, 
                         primerR,
                         minlen, 
                         maxlen,
                         read_q_score,
                         umap_neighbours,
                         hdbscan_neighbours,
                         k,
                         consensus_seq_lim,
                         letter_Q_lim)
    if mode == 'single':

        print('                    O U T P U T   M E R G I N G                    ')
        print('===================================================================')
        
        merge_output(output, mode)

    print('Your results are ready!')
    
    print('Thank you for using PIKE.')
    print('If you find a bug, please report it in a Git issues or contact by email.')

if __name__ == "__main__":
    main()



import numpy as np
import pandas as pd
from os import listdir
from pandas import read_csv, DataFrame

def merge_output(output, mode):

    merged_otu_table = []

    for samp in listdir(f'{output}/results/'):
        
        opn_res = read_csv(f'{output}/results/{samp}/results.tsv', sep='\t', index_col=0)
        count = 0
        merged_otu_table.append(DataFrame(data=opn_res.values, index=opn_res.index, columns=[samp]))
        
        if mode != 'pool':    
            with open(f'{output}/work_dir/{samp}/OTU.fasta', 'w') as opn_fasta:

                for line in opn_res.index:
                    
                    opn_fasta.write(f'>{samp}_{count}_{opn_res["Count"][line]}\n{line}\n')
                    count += 1
        else:
            with open(f'{output}/work_dir/{samp}_OTU.fasta', 'w') as opn_fasta:

                for line in opn_res.index:
                    
                    opn_fasta.write(f'>{samp}_{count}_{opn_res["Count"][line]}\n{line}\n')
                    count += 1

    merged_otu_table = pd.concat(merged_otu_table, axis=1).fillna(0)
    merged_otu_table = merged_otu_table[np.sort(merged_otu_table.columns)].T
    merged_otu_table.to_csv(f'{output}/merged_otu_table.tsv', sep='\t')



import argparse 
import sys
from Pike.src.call_taxonomy import call_taxonomy
from pandas import read_csv

def main():

    parser = argparse.ArgumentParser(description=' P I K E taxonomy caller')

    #______________________Main options_______________________________
    parser.add_argument('-otutab',
                        type=str,
                        help='Path to otu table',
                        required=True,
                        default=None)      
    parser.add_argument('-dbpath',
                        type=str,
                        help='Path to taxonomy DB (SeqID Kingdom;Phylum;Class;Order;Family;Genus;Species)')
    
    parser.add_argument('-identity',
                        type=float,
                        help='Identity treshold (default : 95)',
                        default=95)
    parser.add_argument('-cov_lim',
                        type=float,
                        help='Caverage treshold (default : 60)',
                        default=60)
    parser.add_argument('-eval_lim',
                        type=float,
                        help='E-value treshold (default : 1e-05)',
                        default=60)
    parser.add_argument('-threads',
                        type=int,
                        help='Number of threads (default : 4)',
                        default=4)  
    parser.add_argument('-output',
                        type=str,
                        help='Directory with output files',
                        default='./Pike_results_tax')


    if len(sys.argv)==1:
        parser.print_help(sys.stderr)
        sys.exit(1)

    args = parser.parse_args()
    #Innitializing arguments 
    #__________________________Options__________________________
    otutab = args.otutab
    dbpath = args.dbpath
    identity = args.identity
    cov_lim = args.cov_lim
    eval_lim = args.eval_lim
    output = args.output
    threads = args.threads
    mearged_pike_out = read_csv(otutab, sep='\t', index_col=0)

    tax_table = call_taxonomy(output, 
                             dbpath,
                             mearged_pike_out.T, 
                             threads,
                             identity_filter=identity, 
                             cov_lim=cov_lim, 
                             evalue_filter=eval_lim)
    
    tax_table.to_csv(f'{output}/tax_table.tsv', sep='\t')

if __name__ == "__main__":
    main()



import os
import numpy as np
from os import listdir
from umap import UMAP
from umap.parametric_umap import ParametricUMAP
from hdbscan import HDBSCAN
from subprocess import call
from skbio.stats.composition import clr
from pandas import DataFrame
from multiprocessing import Process
from sklearn.decomposition import PCA
from Pike.src.get_pool import create_pool_dirs, collect_pool_features, run_pool
from Pike.src.get_features import collect_features
from Pike.src.get_filtered_clusters import filter_cluster
from Pike.src.get_good_reads import run_trimming, run_filtering
from Pike.src.get_consensus import get_msa_info, get_consensus, medaka_run, prepare_output
from scipy.sparse import csr_matrix

def create_dirs(output, 
                barcode):

    #workdir dir bases
    os.mkdir(f'{output}/work_dir/{barcode}')
    os.mkdir(f'{output}/results/{barcode}')
    os.mkdir(f'{output}/work_dir/{barcode}/proto_consensus')
    os.mkdir(f'{output}/work_dir/{barcode}/clusters_data_fastq')
    os.mkdir(f'{output}/work_dir/{barcode}/clusters_data_fasta')
    os.mkdir(f'{output}/work_dir/{barcode}/msa')
    os.mkdir(f'{output}/work_dir/{barcode}/medaka')

def run_barcodes(interval,
                 path_to_fastq,
                 output,
                 usereads,
                 trim_primer, 
                 primerF, 
                 primerR,
                 minlen, 
                 maxlen,
                 read_q_score,
                 umap_neighbours,
                 hdbscan_neighbours,
                 k,
                 consensus_seq_lim,
                 letter_Q_lim):
                    
    for barcode in interval:
        
        barcode = barcode.replace('.fastq', '')
        barcode_path = path_to_fastq

        #try:
        create_dirs(output, barcode) #creating and removing directoryes
        
        print(f' R E A D S   P R E P R O C E S S I N G   F O R    :   {barcode} ')
        print('===================================================================')

        if trim_primer == True:
    
            barcode_path = run_trimming(barcode_path, 
                                        barcode, 
                                        output, 
                                        primerF, 
                                        primerR)

        barcode_path = run_filtering(barcode_path, 
                                        barcode, 
                                        output, 
                                        minlen, 
                                        maxlen, 
                                        read_q_score)

        print(f' F E A T U R E S   C O L L E C T I N G   F O R    :   {barcode} ')
        print('===================================================================')

        K_MERS_FREQ, GC_CONTENT, READ_ID, READ_Seq, READ_Q, LENS, QUALITY, BARCODE_ID = collect_features(barcode_path, 
                                                                                            barcode, 
                                                                                            usereads,
                                                                                            k, 
                                                                                            read_q_score) #Features collection
        #___Staged decomposition_________________________________________________________________________________________________________
        print(f' C L U S T E R I N G   S T A G E   F O R          :   {barcode} ')
        print('===================================================================')
        K_MERS_FREQ = np.array(K_MERS_FREQ)
        idx = np.argwhere(np.all(K_MERS_FREQ[..., :] == 0, axis=0))
        K_MERS_FREQ = np.delete(K_MERS_FREQ, idx, axis=1)
        K_MERS_FREQ += 1
        normalize = lambda x: x / np.sum(x)
        K_MERS_FREQ = np.array(list(map(normalize, K_MERS_FREQ)))
        clr_data = clr(K_MERS_FREQ)
        #clr_data = csr_matrix(clr(K_MERS_FREQ.A))
        #print(K_MERS_FREQ.shape)
        import scipy
        #clr_data = clr_data
        pca_model = PCA(n_components=15,
                        random_state=0)#, svd_solver='arpack')
        pca_data = pca_model.fit_transform(clr_data)
       # umap_model = UMAP(n_components=2,
       #                  n_neighbors=umap_neighbours,
       #                  min_dist=0.01,
       #                  metric='euclidean',
       #                  random_state=0)
        umap_model = ParametricUMAP(n_components=2)
        umap_dat = umap_model.fit_transform(pca_data)
        #_______________________________________________________________________________________________________________________________
        RESULT_DICT = {'Read ID' : READ_ID,
                        'BARCODE' : BARCODE_ID, 
                        '1 UMAP COMPONENT' : umap_dat[:, 0], 
                        '2 UMAP COMPONENT' : umap_dat[:, 1], 
                        'Length' : LENS,
                        'READ_Q' : READ_Q,
                        'READ_Seq' : READ_Seq, 
                        'GC content' : GC_CONTENT,
                        'QUALITY' : QUALITY}#,
                        #'K-mers signature' : list(K_MERS_FREQ)}
        RESULT_DF = DataFrame(RESULT_DICT) #All information mearged in df

        #___Cluster identification________________________________________________________________________________________________________
        hdbscan = HDBSCAN(min_cluster_size=hdbscan_neighbours,
                        cluster_selection_epsilon=0.1, 
                        gen_min_span_tree=True,
                        metric='euclidean')
        clusters = hdbscan.fit_predict(umap_dat)
        RESULT_DF['Clusters'] = clusters
        #_______________________________________________________________________________________________________________________________

        RESULT_DF.to_csv(f'{output}/work_dir/{barcode}/metadata_df.tsv', sep='\t')
        clusters = RESULT_DF['Clusters'].unique()
        clusters = np.sort(clusters[clusters != -1])
        filtered_add = {}
        
        #___Cluster processing_____________________________________________________________________________________________________________
        for clt in clusters:
            if clt == -1:
                continue
            clt_dat = RESULT_DF[RESULT_DF['Clusters'] == clt]
            filtered_data = filter_cluster(clt_dat) #filtering clustaer data
            
            if len(filtered_data) < consensus_seq_lim: # can be 0 length for artefact clusters
                continue

            filtered_add[clt] = filtered_data

            #_______Recording cluster data_________________________________________________________________________________________________
                    
            clt_fastq = open(f'{output}/work_dir/{barcode}/clusters_data_fastq/{clt}.fastq', 'w')
            clt_fasta = open(f'{output}/work_dir/{barcode}/clusters_data_fasta/{clt}.fasta', 'w')
            quality_dict = {}
            
            for idx in filtered_data.index:

                read_record = f'@{filtered_data["Read ID"][idx]}\n{filtered_data["READ_Seq"][idx]}\n+\n{filtered_data["READ_Q"][idx]}\n'
                fasta_record = f'>{filtered_data["Read ID"][idx]}\n{filtered_data["READ_Seq"][idx]}\n'
                clt_fastq.write(read_record) #fastq recording
                quality_dict[filtered_data["Read ID"][idx]] = [i for i in RESULT_DF['READ_Q'][idx]] #letter quality adding
                clt_fasta.write(fasta_record) #fasta recording
            
            clt_fastq.close()
            clt_fasta.close()
            #_______________________________________________________________________________________________________________________________
            
            #_______Cluster Consensus Building______________________________________________________________________________________________
                    
            call(f'mafft --quiet {output}/work_dir/{barcode}/clusters_data_fasta/{clt}.fasta > {output}/work_dir/{barcode}/msa/{clt}.msa', shell=True)
            consensus_possitions, possitional_Q_distributions = get_msa_info(output, 
                                                                            barcode, 
                                                                            clt, 
                                                                            quality_dict) #MSA info collection
            protoconsensus = get_consensus(consensus_possitions, 
                                        possitional_Q_distributions, 
                                        letter_Q_lim) #Building a protoconsensus
    
            with open(f'{output}/work_dir/{barcode}/proto_consensus/{clt}.fasta', 'w') as protocons:
    
                protocons.write(f'>clt_{clt}_{len(clt_dat)}\n{protoconsensus}\n') #Recording a protoconsensus    
            #_______________________________________________________________________________________________________________________________

        #_______________________________________________________________________________________________________________________________
        print(f' P O L I S H   C O N S E N S U S    F O R         :   {barcode} ')
        print('===================================================================')

        medaka_run(output, barcode) # run medaka polishing

        print(f' O U T P U T   P R E P A R A T I O N    F O R     :   {barcode} ')
        print('===================================================================')

        prepare_output(output, barcode)

      #  except:
      #      print(f'ERROR WITH {barcode}')

def miltiprocess_analyze(path_to_fastq, 
                         output,
                         mode,
                         threads,
                         usereads,
                         trim_primer,
                         primerF, 
                         primerR,
                         minlen, 
                         maxlen,
                         read_q_score,
                         umap_neighbours,
                         hdbscan_neighbours,
                         k,
                         consensus_seq_lim,
                         letter_Q_lim):
        
    barcodes = listdir(path_to_fastq)
    intervals = np.linspace(0, len(barcodes)+1, threads+1)
    intervals = [barcodes[int(intervals[i]):int(intervals[i+1])] for i in range(len(intervals)-1)] # split to processes 
    procs = []

    if mode == 'single':
        for interval in intervals:
            
            proc = Process(target=run_barcodes, 
                        args=(interval,
                              path_to_fastq,
                              output,
                              usereads,
                              trim_primer, 
                              primerF, 
                              primerR,
                              minlen, 
                              maxlen,
                              read_q_score,
                              umap_neighbours,
                              hdbscan_neighbours,
                              k,
                              consensus_seq_lim,
                              letter_Q_lim))
            proc.daemon = True
            procs.append(proc)
            proc.start()
        
        [proc.join() for proc in procs]
    
    if mode == 'pool':
        
        create_pool_dirs(output)

        for interval in intervals:
            
            proc = Process(target=collect_pool_features, 
                        args=(interval,
                              path_to_fastq,
                              output,
                              usereads,
                              k,
                              trim_primer, 
                              primerF, 
                              primerR,
                              minlen, 
                              maxlen,
                              read_q_score))
            proc.daemon = True
            procs.append(proc)
            proc.start()
        
        [proc.join() for proc in procs]
            
        run_pool(output,
                 threads,
                 umap_neighbours,
                 hdbscan_neighbours,
                 consensus_seq_lim,
                 letter_Q_lim)
        



import numpy as np
from itertools import product 
from Bio.SeqIO import parse
from scipy.sparse import csr_array, vstack


def compress_read(cutted_seq):
    
    new_seq = ''

    for idx in range(len(cutted_seq) - 1):
        if cutted_seq[idx] == cutted_seq[idx + 1]:
            continue

        new_seq += cutted_seq[idx]

    if cutted_seq[idx + 1] != new_seq[-1]:

        new_seq += cutted_seq[idx + 1]
        
    return new_seq 

def get_kmers_signature(seq, k=6):
    
    kmers = {"".join(kmer) : 0 for kmer in list(product("AGTCN", repeat=k))}
 #   step = 1
  #  start = 0
  #  end = k
    reverse_seq = compress_read(seq.reverse_complement())
    forward_seq = compress_read(seq)
    
    #limit = len(forward_seq) - 1

    for start in  range(len(forward_seq) - k):

        end = start + k
        subseq = forward_seq[start: end]
        kmers[subseq] += 1
    #    revers_subseq = reverse_seq[start: end]
    #    kmers[revers_subseq] += 1
    #while end != limit:
        
     #   subseq = forward_seq[start: end]         
     #   kmers[subseq] += 1
        
     #   revers_subseq = reverse_seq[start: end]
     #   kmers[revers_subseq] += 1
        
      #  start, end = start + step, end + step
    
    return kmers

def collect_features(path_to_fastq, 
                     barcode, 
                     usereads, 
                     k,
                     median_Q_lim):

    open_fastq = parse(f'{path_to_fastq}/{barcode}.fastq' , 'fastq')

    K_MERS_FREQ = []
    GC_CONTENT = []
    READ_ID = []
    READ_Seq = []
    READ_Q = []
    LENS = []
    QUALITY = []
    BARCODE_ID = [] #For pool mode
    read_counter = 0
    Q_have = 0
    
    #___Features collection______________________________________________________________
    for seq in open_fastq:
        read_counter += 1
        if Q_have == usereads:
            break
        quallist = np.array(list(map(ord,list(seq.format('fastq').split('\n')[-2]))))-33
        quallist = quallist[quallist < 90]
        median_Q_score = np.median(quallist)
        
        if median_Q_score < median_Q_lim:
            continue
            
        GC_count = str(seq.seq).count('G') + str(seq.seq).count('C')
        
    #    kmer_res = np.array(list(get_kmers_signature(seq.seq, k=k).values()))
      #  kmer_res += 1
      #  kmer_res = kmer_res / np.sum(kmer_res)

#        K_MERS_FREQ.append(csr_array(kmer_res))
        kmer_res = list(get_kmers_signature(seq.seq, k=k).values())

        K_MERS_FREQ.append(kmer_res)
        GC_CONTENT.append(GC_count/ len(seq.seq))
        READ_ID.append(seq.id)
        READ_Seq.append(seq.seq)
        READ_Q.append(seq.format('fastq').split('\n')[3])
        LENS.append(len(seq.seq))
        QUALITY.append(median_Q_score)
        BARCODE_ID.append(barcode)
        Q_have += 1
    #_____________________________________________________________________________________

    print(f'{np.round((1-Q_have/read_counter)*100)}% or reads was dropped') #Добавить log
    print(f'Will be used {len(K_MERS_FREQ)}')
 #   K_MERS_FREQ = vstack(K_MERS_FREQ)

    return K_MERS_FREQ, GC_CONTENT, READ_ID, READ_Seq, READ_Q, LENS, QUALITY, BARCODE_ID

        